"""Model for the position response."""

from __future__ import annotations

from typing import Any, TypedDict


class PositionModel(TypedDict):
    """Model for the position response.

    ref: https://www.traccar.org/api-reference/#tag/Devices/paths/~1devices/get

    WARNING!: The API documentation does not state that null is
    valid for any keys, but this is not the case.
    """

    id: int
    deviceId: int
    protocol: str
    deviceTime: str
    fixTime: str
    serverTime: str
    outdated: bool
    valid: bool
    latitude: float
    geofenceIds: list[int] | None
    longitude: float
    altitude: int
    speed: int
    course: int
    address: str | None
    accuracy: int
    network: dict[str, Any] | None
    attributes: dict[str, Any]



"""Initialize the models module."""

from .device import DeviceModel
from .geofence import GeofenceModel
from .position import PositionModel
from .reports_event import ReportsEventeModel
from .server import ServerModel
from .subscription import SubscriptionData, SubscriptionStatus

__all__ = [
    "DeviceModel",
    "GeofenceModel",
    "PositionModel",
    "ReportsEventeModel",
    "ServerModel",
    "SubscriptionData",
    "SubscriptionStatus",
]



"""Test fixtures and configuration."""

from __future__ import annotations

import logging
from typing import Any, AsyncGenerator

import aiohttp
import pytest
import pytest_asyncio

from pytraccar import ApiClient
from tests.common import MockedRequests, MockResponse, WSMessageHandler

logging.basicConfig(level=logging.ERROR)
logging.getLogger("pytraccar").setLevel(logging.DEBUG)

pytest_plugins = ("pytest_asyncio",)


@pytest.fixture
def mock_requests() -> MockedRequests:
    """Return a new mock request instance."""
    return MockedRequests()


@pytest.fixture
def mock_response() -> MockResponse:
    """Return a new mock response instance."""
    return MockResponse()


@pytest.fixture
def mock_ws_messages() -> WSMessageHandler:
    """Return a new mock ws instance."""
    return WSMessageHandler()


@pytest_asyncio.fixture
async def client_session(
    mock_response: MockResponse,
    mock_requests: MockedRequests,
    mock_ws_messages: WSMessageHandler,
) -> AsyncGenerator[aiohttp.ClientSession, None]:
    """Mock our the request part of the client session."""

    class MockedWSContext:
        @property
        def closed(self) -> bool:
            return len(mock_ws_messages.messages) == 0

        async def receive(self) -> Any:
            return mock_ws_messages.get()

        async def __aenter__(self) -> MockedWSContext:
            return self

        async def __aexit__(self, *args: Any) -> None:
            pass

        def __aiter__(self) -> MockedWSContext:
            return self

        async def __anext__(self) -> Any:
            if len(mock_ws_messages.messages) == 0:
                raise StopAsyncIteration
            return mock_ws_messages.get()

    async def _mocked_ws_connect(*_: Any, **__: Any) -> Any:
        return MockedWSContext()

    async def _mocked_request(*args: Any, **kwargs: Any) -> Any:
        if len(args) > 2:
            mock_response.mock_endpoint = args[2].split("/api/")[-1]
            mock_requests.add({"method": args[1], "url": args[2], **kwargs})
        else:
            mock_response.mock_endpoint = args[1].split("/api/")[-1]
            mock_requests.add({"method": args[0], "url": args[1], **kwargs})
        return mock_response

    async with aiohttp.ClientSession() as session:
        mock_requests.clear()
        session._request = _mocked_request  # noqa: SLF001
        session._ws_connect = _mocked_ws_connect  # noqa: SLF001
        yield session


@pytest_asyncio.fixture
async def api_client(
    client_session: AsyncGenerator[aiohttp.ClientSession, None],
) -> AsyncGenerator[ApiClient, None]:
    """Fixture to provide a API Client."""
    yield ApiClient(
        host="127.0.0.1",
        port=1337,
        token="test",
        client_session=client_session,
    )



"""Model for the server response."""

from __future__ import annotations

from typing import Any, TypedDict


class ServerModel(TypedDict):
    """Model for the server response.

    ref: https://www.traccar.org/api-reference/#tag/Server/paths/~1server/get

    WARNING!: The API documentation does not state that null is
    valid for any keys, but this is not the case.
    """

    id: int
    registration: bool
    readonly: bool
    deviceReadonly: bool
    limitCommands: bool
    map: str | None
    bingKey: str | None
    mapUrl: str | None
    poiLayer: str | None
    latitude: float
    longitude: float
    zoom: int
    twelveHourFormat: bool
    version: str
    forceSettings: bool
    coordinateFormat: str | None
    attributes: dict[str, Any]
    openIdEnabled: bool
    openIdForce: bool



"""Test API endpoint."""

import pytest

from pytraccar import ApiClient


@pytest.mark.asyncio
async def test_server(api_client: ApiClient) -> None:
    """Test /server endpoint."""
    response = await api_client.get_server()
    assert response["id"] == 0


@pytest.mark.asyncio
async def test_devices(api_client: ApiClient) -> None:
    """Test /devices endpoint."""
    response = await api_client.get_devices()
    assert isinstance(response, list)
    assert response[0]["id"] == 0


@pytest.mark.asyncio
async def test_geofences(api_client: ApiClient) -> None:
    """Test /geofences endpoint."""
    response = await api_client.get_geofences()
    assert isinstance(response, list)
    assert response[0]["id"] == 0


@pytest.mark.asyncio
async def test_positions(api_client: ApiClient) -> None:
    """Test /positions endpoint."""
    response = await api_client.get_positions()
    assert isinstance(response, list)
    assert response[0]["id"] == 0


@pytest.mark.asyncio
async def test_reports_events(api_client: ApiClient) -> None:
    """Test /reports/events endpoint."""
    response = await api_client.get_reports_events()
    assert isinstance(response, list)
    assert response[0]["id"] == 0



"""API client for Traccar servers."""

from __future__ import annotations

import asyncio
from contextlib import suppress
from datetime import UTC, datetime, timedelta
from logging import Logger, getLogger
from typing import TYPE_CHECKING, Any

import aiohttp

from .exceptions import (
    TraccarAuthenticationException,
    TraccarConnectionException,
    TraccarException,
    TraccarResponseException,
)
from .models import (
    SubscriptionData,
    SubscriptionStatus,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from .models import (
        DeviceModel,
        GeofenceModel,
        PositionModel,
        ReportsEventeModel,
        ServerModel,
    )


_LOGGER: Logger = getLogger(__package__)


class ApiClient:
    """Class for interacting with the Traccar API."""

    def __init__(
        self,
        host: str,
        token: str,
        client_session: aiohttp.ClientSession,
        *,
        port: int | None = None,
        ssl: bool = False,
        verify_ssl: bool = True,
        ws_heartbeat: int = 120,
        **_: Any,
    ) -> None:
        """Initialize the API client."""
        self._base_url = f"http{'s' if ssl else ''}://{host}:{port or 8082}/api"
        self._client_session = client_session
        self._token = token
        self._verify_ssl = verify_ssl
        self._subscription_status = SubscriptionStatus.DISCONNECTED
        self._ws_heartbeat = ws_heartbeat

    @property
    def subscription_status(self) -> SubscriptionStatus:
        """Return the current subscription status."""
        return self._subscription_status

    async def _call_api(
        self,
        endpoint: str,
        method: str = "GET",
        *,
        params: list[tuple[str, str | int]] | None = None,
        headers: dict[str, str] | None = None,
        data: Any | None = None,
        **_: Any,
    ) -> Any:
        """Call the API endpoint and return the response."""
        try:
            async with self._client_session.request(
                method=method,
                url=f"{self._base_url}/{endpoint}",
                ssl=self._verify_ssl,
                params=params,
                data=data,
                headers={
                    aiohttp.hdrs.ACCEPT: "application/json",
                    aiohttp.hdrs.AUTHORIZATION: f"Bearer {self._token}",
                    aiohttp.hdrs.CONTENT_TYPE: "application/json",
                    **(headers or {}),
                },
                timeout=aiohttp.ClientTimeout(total=10),
            ) as response:
                if response.status == 401:
                    raise TraccarAuthenticationException("Unauthorized")
                if response.status == 200:
                    return await response.json()

                raise TraccarResponseException(f"{response.status}: {response.reason}")
        except (TraccarAuthenticationException, TraccarResponseException):
            raise
        except TimeoutError as exception:
            raise TraccarConnectionException(
                "Timeouterror connecting to Traccar"
            ) from exception
        except (aiohttp.ClientError, asyncio.CancelledError) as exception:
            raise TraccarConnectionException(
                f"Could not communicate with Traccar - {exception}"
            ) from exception
        except Exception as exception:  # pylint: disable=broad-except
            raise TraccarException(f"Unexpected error - {exception}") from exception

    async def get_server(self) -> ServerModel:
        """Get server information."""
        response: ServerModel = await self._call_api("server")
        return response

    async def get_devices(self) -> list[DeviceModel]:
        """Get all devices from the Traccar API."""
        response: list[DeviceModel] = await self._call_api("devices")
        return response

    async def get_geofences(self) -> list[GeofenceModel]:
        """Get all geofences from the Traccar API."""
        response: list[GeofenceModel] = await self._call_api("geofences")
        return response

    async def get_positions(self) -> list[PositionModel]:
        """Get all positions from the Traccar API."""
        response: list[PositionModel] = await self._call_api("positions")
        return response

    async def get_reports_events(
        self,
        *,
        devices: list[int] | None = None,
        groups: list[int] | None = None,
        event_types: list[str] | None = None,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
        **_: Any,
    ) -> list[ReportsEventeModel]:
        """Get events."""
        datetime_now = datetime.now(tz=UTC).replace(tzinfo=None)
        start_time = start_time or datetime_now
        end_time = end_time or datetime_now - timedelta(hours=30)
        response: list[ReportsEventeModel] = await self._call_api(
            "reports/events",
            params=[
                ("to", start_time.isoformat() + "Z"),
                ("from", end_time.isoformat() + "Z"),
                *[("deviceId", device) for device in devices or []],
                *[("groupId", group) for group in groups or []],
                *[("type", value) for value in event_types or ""],
            ],
        )
        return response

    async def subscribe(
        self, callback: Callable[[SubscriptionData], Awaitable[None]]
    ) -> None:
        """Subscribe to events."""

        async def _subscriber() -> None:
            self._subscription_status = SubscriptionStatus.CONNECTING
            async with self._client_session.ws_connect(
                url=f"{self._base_url}/socket",
                verify_ssl=self._verify_ssl,
                heartbeat=self._ws_heartbeat,
            ) as ws:
                self._subscription_status = SubscriptionStatus.CONNECTED
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        if not (data := msg.json()):
                            # Ignore empty messages
                            continue
                        try:
                            await callback(
                                {
                                    "devices": None,
                                    "events": None,
                                    "positions": None,
                                    **data,
                                }
                            )
                        except Exception as exception:
                            _LOGGER.exception(
                                "Exception while handling message: %s(%s)",
                                exception.__class__.__name__,
                                exception,
                            )
                    elif msg.type in (
                        aiohttp.WSMsgType.CLOSE,
                        aiohttp.WSMsgType.CLOSED,
                        aiohttp.WSMsgType.CLOSING,
                        aiohttp.WSMsgType.ERROR,
                    ):
                        raise TraccarConnectionException(
                            f"WebSocket connection closed with {msg.type.name}"
                        )
                    else:
                        _LOGGER.warning("Unexpected message type %s", msg.type.name)

        try:
            # https://www.traccar.org/api-reference/#tag/Session/paths/~1session/post
            await self._call_api(
                f"session?token={self._token}",
                method="GET",
                headers={
                    aiohttp.hdrs.CONTENT_TYPE: "application/x-www-form-urlencoded",
                },
            )
            await _subscriber()
        except asyncio.CancelledError:
            self._subscription_status = SubscriptionStatus.DISCONNECTED
        except Exception as exception:  # pylint: disable=broad-except
            self._subscription_status = SubscriptionStatus.ERROR
            if isinstance(exception, TraccarConnectionException):
                raise
            if isinstance(exception, asyncio.TimeoutError):
                raise TraccarConnectionException(
                    "Timeout error connecting to Traccar"
                ) from exception
            if isinstance(exception, aiohttp.ClientError):
                raise TraccarConnectionException(
                    "Could not communicate with Traccar"
                ) from exception
            raise TraccarException("Unexpected error") from exception
        finally:
            # Close the session if we can
            # https://www.traccar.org/api-reference/#tag/Session/paths/~1session/delete
            with suppress(TraccarException):
                await self._call_api(
                    "session",
                    method="DELETE",
                    headers={},
                )



"""[GitHub repository](https://github.com/ludeeus/pytraccar)."""

from .client import ApiClient
from .exceptions import (
    TraccarAuthenticationException,
    TraccarConnectionException,
    TraccarException,
    TraccarResponseException,
)
from .models import (
    DeviceModel,
    GeofenceModel,
    PositionModel,
    ReportsEventeModel,
    ServerModel,
    SubscriptionData,
    SubscriptionStatus,
)

__all__ = [
    "ApiClient",
    "DeviceModel",
    "GeofenceModel",
    "PositionModel",
    "ReportsEventeModel",
    "ServerModel",
    "SubscriptionData",
    "SubscriptionStatus",
    "TraccarAuthenticationException",
    "TraccarConnectionException",
    "TraccarException",
    "TraccarResponseException",
]



"""Common helpers for tests."""

from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from aiohttp import WSMsgType


class WSMessage:
    """WSMessage."""

    def __init__(self, messagetype: WSMsgType, json: dict | None = None) -> None:
        """Initialize."""
        self.type = messagetype
        self._json = json

    def json(self) -> dict | None:
        """json."""
        return self._json


def load_response(filename: str) -> dict[str, Any]:
    """Load a response."""
    filename = f"{filename.split('?')[0]}.json" if "." not in filename else filename
    path = Path(
        Path.resolve(Path(__file__)).parent,
        "responses",
        filename.lower().replace("/", "_"),
    )
    with path.open(encoding="utf-8") as fptr:
        return json.loads(fptr.read())


class WSMessageHandler:
    """WSMessageHandler."""

    def __init__(self) -> None:
        """Initialize."""
        self.messages = []

    def add(self, msg: WSMessage) -> None:
        """Add."""
        self.messages.append(msg)

    def get(self) -> WSMessage:
        """Get."""
        return (
            self.messages.pop(0)
            if self.messages
            else WSMessage(messagetype=WSMsgType.CLOSED)
        )


@dataclass
class MockResponse:
    """Mock response class."""

    _count = 0

    mock_data: Any | None = None
    mock_data_list: list[Any] | None = None
    mock_endpoint: str = ""
    mock_headers: dict[str, str] | None = None
    mock_raises: BaseException | None = None
    mock_status: int = 200
    _in_context: bool = False

    async def __aenter__(self) -> MockResponse:
        self._in_context = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self._in_context = False
        self.release()
        await self.wait_for_close()

    @property
    def status(self) -> int:
        """status."""
        return self.mock_status

    @property
    def reason(self) -> str:
        """Return the reason."""
        return "unknown"

    async def json(self, **_: Any) -> Any:
        """json."""
        if self.mock_raises is not None:
            raise self.mock_raises  # pylint: disable=raising-bad-type
        if self.mock_data_list:
            data = self.mock_data_list[self._count]
            self._count += 1
            return data
        if self.mock_data is not None:
            return self.mock_data
        return load_response(self.mock_endpoint)

    def release(self) -> None:
        """release."""

    def clear(self) -> None:
        """clear."""
        self.mock_data = None
        self.mock_endpoint = ""
        self.mock_headers = None
        self.mock_raises = None
        self.mock_status = 200

    async def wait_for_close(self) -> None:
        """wait_for_close."""


class MockedRequests:
    """Mock request class."""

    def __init__(self) -> None:
        """Initialize."""
        self._calls = []

    def add(self, url: str) -> None:
        """add."""
        self._calls.append(url)

    def clear(self) -> None:
        """clear."""
        self._calls.clear()

    def __repr__(self) -> str:
        """repr."""
        return f"<MockedRequests: {self._calls}>"

    @property
    def called(self) -> int:
        """count."""
        return len(self._calls)

    def has(self, string: str) -> bool:
        """has."""
        return bool([entry for entry in self._calls if string in entry])

    @property
    def last_request(self) -> MockResponse:
        """Last url."""
        return self._calls[-1]



"""Model for the subscription."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .device import DeviceModel
    from .position import PositionModel
    from .reports_event import ReportsEventeModel


class SubscriptionStatus(StrEnum):
    """Model for the subscription status."""

    CONNECTED = "connected"
    CONNECTING = "connecting"
    DISCONNECTED = "disconnected"
    ERROR = "error"


class SubscriptionData(TypedDict):
    """Model for the subscription data.

    ref: https://www.traccar.org/traccar-api/
    """

    devices: list[DeviceModel] | None
    positions: list[PositionModel] | None
    events: list[ReportsEventeModel] | None



"""Test subscription."""

from __future__ import annotations

import asyncio
from typing import Any, NoReturn
from unittest.mock import patch

import aiohttp
import pytest
from aiohttp import WSMsgType

from pytraccar import (
    ApiClient,
    SubscriptionStatus,
    TraccarConnectionException,
    TraccarException,
)
from tests.common import WSMessage, WSMessageHandler


@pytest.mark.parametrize(
    "messages",
    [
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={}),
            WSMessage(messagetype=WSMsgType.TEXT, json=None),
        ],
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={"devices": []}),
        ],
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={"positions": []}),
        ],
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={"events": []}),
        ],
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={"devices": []}),
            WSMessage(messagetype=WSMsgType.TEXT, json={"positions": []}),
            WSMessage(messagetype=WSMsgType.TEXT, json={"events": []}),
        ],
        [
            WSMessage(messagetype=WSMsgType.TEXT, json={"events": [], "devices": []}),
        ],
    ],
)
@pytest.mark.asyncio
async def test_subscription_text_message(
    api_client: ApiClient,
    messages: list[WSMessage],
    mock_ws_messages: WSMessageHandler,
) -> None:
    """Test subscription text message."""
    _handled = []
    _expected_handled = []
    for message in messages:
        mock_ws_messages.add(message)
        if message.type == WSMsgType.TEXT and (data := message.json()):
            _expected_handled.append(
                {
                    "devices": None,
                    "events": None,
                    "positions": None,
                    **data,
                }
            )

    async def _handler(data: Any) -> None:
        _handled.append(data)

    await api_client.subscribe(_handler)
    assert _handled == _expected_handled


@pytest.mark.parametrize(
    "message",
    [
        WSMessage(messagetype=WSMsgType.CLOSE),
        WSMessage(messagetype=WSMsgType.CLOSED),
        WSMessage(messagetype=WSMsgType.CLOSING),
        WSMessage(messagetype=WSMsgType.ERROR),
    ],
)
@pytest.mark.asyncio
async def test_subscription_stopping_message(
    api_client: ApiClient,
    message: WSMessage,
    mock_ws_messages: WSMessageHandler,
) -> None:
    """Test subscription stopping message."""
    _handled = []
    mock_ws_messages.add(message)

    async def _handler(data: Any) -> None:
        _handled.append(data)

    with pytest.raises(
        TraccarConnectionException,
        match=f"WebSocket connection closed with {message.type.name}",
    ):
        await api_client.subscribe(_handler)
    assert len(_handled) == 0


@pytest.mark.parametrize(
    "message",
    [
        WSMessage(messagetype=WSMsgType.CONTINUATION),
        WSMessage(messagetype=WSMsgType.BINARY),
        WSMessage(messagetype=WSMsgType.PING),
        WSMessage(messagetype=WSMsgType.PONG),
    ],
)
@pytest.mark.asyncio
async def test_subscription_unknown_type(
    api_client: ApiClient,
    message: WSMessage,
    mock_ws_messages: WSMessageHandler,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test subscription unknown type."""
    _handled = []
    mock_ws_messages.add(message)

    async def _handler(data: Any) -> None:
        _handled.append(data)

    assert f"Unexpected message type {message.type.name}" not in caplog.text

    await api_client.subscribe(_handler)

    assert len(_handled) == 0
    assert f"Unexpected message type {message.type.name}" in caplog.text


@pytest.mark.asyncio
async def test_subscription_bad_handler(
    api_client: ApiClient,
    mock_ws_messages: WSMessageHandler,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test subscription unknown type."""
    mock_ws_messages.add(WSMessage(messagetype=WSMsgType.TEXT, json={"devices": []}))

    async def _handler(_: Any) -> NoReturn:
        raise ValueError("Bad handler")

    await api_client.subscribe(_handler)

    assert "Exception while handling message: ValueError(Bad handler)" in caplog.text


@pytest.mark.parametrize(
    ("side_effect", "raises", "with_message"),
    [
        (
            KeyError("boom"),
            TraccarException,
            "Unexpected error",
        ),
        (
            asyncio.TimeoutError(),
            TraccarConnectionException,
            "Timeout error connecting to Traccar",
        ),
        (
            aiohttp.ClientError("boom"),
            TraccarConnectionException,
            "Could not communicate with Traccar",
        ),
        (
            TraccarConnectionException(),
            TraccarConnectionException,
            None,
        ),
    ],
)
@pytest.mark.asyncio
async def test_subscription_exceptions(
    api_client: ApiClient,
    side_effect: Exception,
    raises: Exception,
    with_message: str | None,
) -> None:
    """Test subscription exceptions."""
    assert api_client.subscription_status == SubscriptionStatus.DISCONNECTED
    with patch("aiohttp.ClientSession.ws_connect", side_effect=side_effect):
        if with_message is not None:
            with pytest.raises(raises, match=with_message):
                await api_client.subscribe(None)
        else:
            with pytest.raises(raises):
                await api_client.subscribe(None)

    assert api_client.subscription_status == SubscriptionStatus.ERROR


@pytest.mark.asyncio
async def test_subscription_cancelation(api_client: ApiClient) -> None:
    """Test subscription exceptions."""
    assert api_client.subscription_status == SubscriptionStatus.DISCONNECTED
    with patch(
        "aiohttp.ClientSession.ws_connect", side_effect=asyncio.CancelledError("boom")
    ):
        await api_client.subscribe(None)

    assert api_client.subscription_status == SubscriptionStatus.DISCONNECTED



"""Base client tests."""

import pytest
from aiohttp import ClientSession

from pytraccar import ApiClient


@pytest.mark.asyncio
async def test_client_init(client_session: ClientSession) -> None:
    """Test client init."""
    client_params = {
        "host": "127.0.0.1",
        "token": "test",
        "port": 8080,
        "client_session": client_session,
    }

    assert ApiClient(**client_params)._base_url == "http://127.0.0.1:8080/api"  # noqa: SLF001
    assert (
        ApiClient(**{**client_params, "port": None})._base_url  # noqa: SLF001
        == "http://127.0.0.1:8082/api"
    )
    assert (
        ApiClient(**{**client_params, "ssl": True})._base_url  # noqa: SLF001
        == "https://127.0.0.1:8080/api"
    )



"""Model for the geofence response."""

from __future__ import annotations

from typing import Any, TypedDict


class GeofenceModel(TypedDict):
    """Model for the geofence response.

    ref: https://www.traccar.org/api-reference/#tag/Geofences/paths/~1geofences/get

    WARNING!: The API documentation does not state that null is
    valid for any keys, but this is not the case.
    """

    id: int
    name: str
    description: str | None
    area: str
    calendarId: str
    attributes: dict[str, Any]



"""Model for the reports/events response."""

from __future__ import annotations

from typing import Any, TypedDict


class ReportsEventeModel(TypedDict):
    """Model for the reports/events response.

    ref: https://www.traccar.org/api-reference/#tag/Reports/paths/~1reports~1events/get

    WARNING!: The API documentation does not state that null is
    valid for any keys, but this is not the case.
    """

    id: int
    type: str
    eventTime: str
    deviceId: int
    positionId: int
    geofenceId: int
    maintenanceId: int
    attributes: dict[str, Any]



"""Example usage of pytraccar."""

import asyncio
import logging
import os

import aiohttp

from pytraccar import ApiClient

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[logging.StreamHandler()],
)


async def test() -> None:
    """Example usage of pytraccar."""
    async with aiohttp.ClientSession(
        cookie_jar=aiohttp.CookieJar(unsafe=True)
    ) as client_session:
        client = ApiClient(
            host=os.environ["TRACCAR_HOST"],
            port=os.environ.get("TRACCAR_PORT", 8082),
            token=os.environ["TRACCAR_TOKEN"],
            client_session=client_session,
        )
        server = await client.get_server()
        logging.info(
            "Connected to Traccar server (%s:%s) which is running version %s",
            os.environ["TRACCAR_HOST"],
            os.environ.get("TRACCAR_PORT", 8082),
            server["version"],
        )


asyncio.run(test())



"""Custom exceptions for pytraccar."""


class TraccarException(Exception):
    """Base class for all pytraccar exceptions."""


class TraccarAuthenticationException(TraccarException):
    """Exception for authentication errors."""


class TraccarConnectionException(TraccarException):
    """Exception for connection errors."""


class TraccarResponseException(TraccarException):
    """Exception for response errors."""



"""Model for the devices response."""

from __future__ import annotations

from typing import Any, TypedDict


class DeviceModel(TypedDict):
    """Model for the devices response.

    ref: https://www.traccar.org/api-reference/#tag/Devices/paths/~1devices/get

    WARNING!: The API documentation does not state that null is
    valid for any keys, but this is not the case.
    """

    id: int
    name: str
    uniqueId: str
    status: str
    disabled: bool
    lastUpdate: str | None
    positionId: int
    groupId: int
    phone: str | None
    model: str | None
    contact: str | None
    category: str | None
    attributes: dict[str, Any]



"""Base client tests."""

import asyncio

import aiohttp
import pytest

from pytraccar import (
    ApiClient,
    TraccarAuthenticationException,
    TraccarConnectionException,
    TraccarException,
    TraccarResponseException,
)
from tests.common import MockResponse


@pytest.mark.asyncio
async def test_base_api(api_client: ApiClient) -> None:
    """Test base API."""
    response = await api_client.get_server()
    assert response["bingKey"] == "string"


@pytest.mark.asyncio
async def test_base_api_unauthenticated(
    api_client: ApiClient, mock_response: MockResponse
) -> None:
    """Test unauthenticated base API."""
    mock_response.mock_status = 401
    with pytest.raises(TraccarAuthenticationException):
        await api_client.get_server()


@pytest.mark.asyncio
async def test_base_api_issue(
    api_client: ApiClient, mock_response: MockResponse
) -> None:
    """Test API issue."""
    mock_response.mock_status = 500
    with pytest.raises(TraccarResponseException):
        await api_client.get_server()


@pytest.mark.asyncio
async def test_base_api_timeout(
    api_client: ApiClient, mock_response: MockResponse
) -> None:
    """Test API issue."""
    mock_response.mock_raises = asyncio.TimeoutError
    with pytest.raises(TraccarConnectionException):
        await api_client.get_server()

    mock_response.mock_raises = aiohttp.ClientError
    with pytest.raises(TraccarConnectionException):
        await api_client.get_server()

    mock_response.mock_raises = TypeError
    with pytest.raises(TraccarException):
        await api_client.get_server()
